package no.statnett.ecp.brs;

import no.statnett.ecp.brs.actions.*;
import no.statnett.ecp.brs.state.*;
import no.statnett.ecp.utils.EcpBroker;
import no.statnett.ecp.utils.Options;

import java.io.File;
import java.io.IOException;
import java.nio.file.Files;
import java.security.KeyManagementException;
import java.security.NoSuchAlgorithmException;
import java.util.*;
import java.util.stream.Collectors;

public class BrokerShell {

    public static final String VERSION = "v1.4.14"; //

    // Options from input-args
    private static Map<String, Login> extraLogins = new HashMap<>();
    private static List<String> commands = null;

    // Useful global variables, set based on commands
    private static boolean echo = true;
    private static final QueueFilter queueFilter = new QueueFilter();
    private static final MessageFilter messageFilter = new MessageFilter();
    private static final List<String> filterHistory = new ArrayList<>();
    private static boolean filterChangedViewNotUpdated = true;
    private static boolean messageFilterChangedViewNotUpdated = true;

    private static List<String> parseOptions(String[] initialArgs) throws IOException {
        List<String> initArgsList = new ArrayList<>(Arrays.asList(initialArgs));
        String commandsFromFile = Options.parseString(initArgsList, "-f");
        String commandsFromArg = Options.parseString(initArgsList, "-c");
        if (commandsFromFile != null) {
            File commandFile = Options.checkFile(commandsFromFile, false);
            commands = Files.readAllLines(commandFile.toPath());
        } else if (commandsFromArg != null) {
            commands = Arrays.asList(commandsFromArg.split(";"));
        }
        if (commands != null) { // Some commands have been specified, we'll filter out empty lines and comments
            commands = commands.stream().filter(c -> !c.trim().isEmpty()).filter(c -> !c.trim().startsWith("#")).collect(Collectors.toList());
            if (commands.isEmpty()) {
                System.out.println("Error: No commands found in file (-f) or in argument, exiting");
                System.exit(1);
            }
        }
        String loginFileStr = Options.parseString(initArgsList, "-h");
        if (loginFileStr != null) {
            File loginFile = Options.checkFile(loginFileStr, false);
            extraLogins = parseLoginFile(loginFile);
        }
        return initArgsList;
    }


    public static void main(String[] args) throws IOException, NoSuchAlgorithmException, KeyManagementException {
        // READ OPTIONS
        List<String> mandatoryArgs = parseOptions(args);
        if (mandatoryArgs.size() != 1) {
            if (!mandatoryArgs.isEmpty()) {
                System.out.println("\nError: Expected exactly 1 (one) mandatory argument, either URL/PATH to broker, but received " + mandatoryArgs.size() + "\n");
            }
            if (echo)
                usage();
            System.exit(1);
        }

        // LOAD CONFIG - PREPARE INITIAL STATE
        Map<String, String> brokerConfig = EcpBroker.findBrokerConfig(mandatoryArgs.get(0), false);
        if (!extraLogins.isEmpty()) {
            extraLogins.remove(brokerConfig.get("host")); // Remove the main host from the extraHosts list - to avoid double listing
        }
        Config config = new Config(brokerConfig, extraLogins);
        State state = new State();

        // RUN
        if (commands != null) {
            script(state, config, commands);
        } else {
            interactive(state, config);
        }
    }

    private static void executeCommand(State state, Config config, String input) throws NoSuchAlgorithmException, IOException, KeyManagementException {
        if (input.startsWith("e"))
            echo = !input.substring(1).trim().equalsIgnoreCase("off");
        else if (echo && input.equalsIgnoreCase("h"))
            help();
        else if (echo && input.equalsIgnoreCase("y"))
            showSettings();
        else if (echo && input.equalsIgnoreCase("f"))
            filterHistory();
        else if (input.equalsIgnoreCase("v"))
            state.setQiList(refreshQueueList(config));
        else if (input.startsWith("age"))
            setAgeFilter(input);
        else if (input.startsWith("type"))
            setTypeFilter(input);
        else if (input.startsWith("baid"))
            setBaMessageIdFilter(input);
        else if (input.startsWith("snd"))
            setSendFilter(input);
        else if (input.startsWith("rcv"))
            setReceiveFilter(input);
        else if (input.startsWith("c"))
            setConsumerFilter(input);
        else if (input.startsWith("z"))
            setSizeFilter(input);
        else if (input.startsWith("s"))
            setSearchFilter(input);
        else if (input.startsWith("q"))
            setQueuenameFilter(input);
        else if (input.startsWith("a"))
            setAscendingSort(input);
        else if (input.startsWith("d"))
            setDescendingSort(input);
        else if (input.startsWith("l"))
            setLimitFilter(input);
        else { // The commands below will perform an action - we'll prevent that action if v-command has not been run
            if (filterChangedViewNotUpdated)
                System.out.println("Run 'v' command, then try again. Filters may have changed or no queues have been listed yet");
            else if (input.startsWith("p"))
                PurgeQueue.purge(state.getQiList(), config, input, echo);
            else if (input.startsWith("r"))
                RemoveMessages.remove(state.getQiList(), config, input, echo, messageFilter);
            else if (input.startsWith("n"))
                DeleteQueue.delete(state.getQiList(), config, input, echo);
            else if (input.startsWith("m"))
                listMessages(state, config, input, echo);
            else if (input.startsWith("j"))
                printQueueInfoToJsonFile(state.getQiList(), input);
        }
    }

    private static void script(State state, Config config, List<String> commmands) throws NoSuchAlgorithmException, IOException, KeyManagementException {
        // There is at least ONE command - this has been checked in parseOptions()
        boolean echoOn = !commmands.get(0).startsWith("e") || !commmands.get(0).endsWith("off");
        for (String command : commmands) {
            if (echoOn)
                System.out.println(">" + command);
            executeCommand(state, config, command);
            echoOn = echo; // The executeCommand will update the echo-flag if the command is "e off" or "e on"
            if (command.equalsIgnoreCase("x"))
                break;
        }
        if (echo)
            System.out.println("Exiting BrokerShell");
    }

    private static void interactive(State state, Config config) throws NoSuchAlgorithmException, IOException, KeyManagementException {
        System.out.println("BrokerShell for Artemis v" + VERSION + " - Interactive mode. Enter h for help");
        Scanner scanner = new Scanner(System.in);
        String input;
        do {
            System.out.print(">");
            input = scanner.nextLine();
            if (!input.trim().isEmpty()) {
                executeCommand(state, config, input);
            }
        } while (!input.equalsIgnoreCase("x"));
        scanner.close();
        System.out.println("Exiting BrokerShell");
    }

    private static void listMessages(State state, Config config, String input, boolean echo) {
        Integer tmp = Div.getInt(input, false);
        List<MessageInfo> miList = new ArrayList<>();
        boolean miListRefreshed = false;
        if (tmp == null || tmp < 0) {
            miList = ListMessages.Retrieve.retrieveAndFilter(state.getQiList(), config, messageFilter);
            miListRefreshed = true;
        } else if (tmp > 0 && tmp <= state.getQiList().size()) {
            List<QueueInfo> selectedQIList = new ArrayList<>();
            selectedQIList.add(state.getQiList().get(tmp - 1));
            miList = ListMessages.Retrieve.retrieveAndFilter(selectedQIList, config, messageFilter);
            miListRefreshed = true;
        } else {
            System.out.println("Invalid queue number: " + tmp + ", command ignored");
        }
        if (miListRefreshed) {
            state.setMiList(miList);
            ListMessages.Print.printMI(state.getMiList(), echo);
            messageFilterChangedViewNotUpdated = false;
        }
    }

    private static void setSizeFilter(String input) {
        Integer tmp = Div.getInt(input);
        if (tmp == null)
            return;
        if (tmp < 1) {
            queueFilter.setMinQueueSize(0);
            if (echo)
                System.out.println("Minimum size limit is turned off, will show queues of all sizes");
        } else {
            queueFilter.setMinQueueSize(tmp);
            if (echo)
                System.out.println("Minimum size of queues is set to " + tmp);
        }
        filterChangedViewNotUpdated = true;
    }

    private static void setLimitFilter(String input) {
        Integer tmp = Div.getInt(input);
        if (tmp == null)
            return;
        if (tmp < 1) {
            queueFilter.setMaxQueueListSize(0);
            if (echo)
                System.out.println("Limit is turned off, will show all queues that matches the filter-settings");
        } else {
            queueFilter.setMaxQueueListSize(tmp);
            if (echo)
                System.out.println("Limit is set to " + tmp + " queues");
        }
        filterChangedViewNotUpdated = true;
    }

    private static void setAgeFilter(String input) {
        Integer tmp = Div.getInt(input);
        if (tmp == null)
            return;
        if (tmp < 1) {
            messageFilter.setMinAgeMinutes(0);
            if (echo)
                System.out.println("Limit is turned off, will show all message regardless of age");
        } else {
            messageFilter.setMinAgeMinutes(tmp);
            if (echo)
                System.out.println("Limit is set to minimum " + tmp + " minutes old messages");
        }
        messageFilterChangedViewNotUpdated = true;
    }

    private static void setTypeFilter(String input) {
        String tmp = input.substring(4).trim();
        if (tmp.isEmpty()) {
            messageFilter.setMessageType(null);
            if (echo)
                System.out.println("Filtering by message/business-type is turned off, will match ALL message types");
        } else {
            messageFilter.setMessageType(tmp);
            if (echo)
                System.out.println("Filtering by message/business-type is set to " + tmp);
        }
        messageFilterChangedViewNotUpdated = true;
    }

    private static void setBaMessageIdFilter(String input) {
        String tmp = input.substring(4).trim();
        if (tmp.isEmpty()) {
            messageFilter.setBaMessageId(null);
            if (echo)
                System.out.println("Filtering by BAMessageId is turned off, will match ALL BAMessageIds");
        } else {
            messageFilter.setBaMessageId(tmp);
            if (echo)
                System.out.println("Filtering by BAMessageId is set to " + tmp);
        }
        messageFilterChangedViewNotUpdated = true;
    }

    private static void setSendFilter(String input) {
        String tmp = input.substring(3).trim();
        if (tmp.isEmpty()) {
            messageFilter.setSender(null);
            if (echo)
                System.out.println("Filtering by sender is turned off, will match ALL senders");
        } else {
            messageFilter.setSender(tmp);
            if (echo)
                System.out.println("Filtering by sender is set to " + tmp);
        }
        messageFilterChangedViewNotUpdated = true;
    }

    private static void setReceiveFilter(String input) {
        String tmp = input.substring(3).trim();
        if (tmp.isEmpty()) {
            messageFilter.setReceiver(null);
            if (echo)
                System.out.println("Filtering by receiver is turned off, will match ALL receivers");
        } else {
            messageFilter.setReceiver(tmp);
            if (echo)
                System.out.println("Filtering by receiver is set to " + tmp);
        }
        messageFilterChangedViewNotUpdated = true;
    }




    private static void setDescendingSort(String input) {
        Integer tmp = Div.getInt(input);
        if (tmp == null)
            return;
        if (tmp > 6) {
            System.out.println("Sorting by column " + tmp + " is not supported, only 1-6, command is ignored");
        } else if (tmp < 1) {
            if (echo) {
                if (queueFilter.getSortAscending() > 0) {
                    System.out.println("Descending sort is turned off, but has no effect since ascending sort of column " + queueFilter.getSortAscending() + " takes precedence");
                } else {
                    System.out.println("Descending sort is turned off, will show default listing");
                }
            }
            queueFilter.setSortDescending(0);
        } else {
            queueFilter.setSortDescending(tmp);
            queueFilter.setSortAscending(0);
            if (echo)
                System.out.println("Sorting by column " + tmp + " in descending order");
        }
    }

    private static void setAscendingSort(String input) {
        Integer tmp = Div.getInt(input);
        if (tmp == null)
            return;
        if (tmp > 6) {
            System.out.println("Sorting by column " + tmp + " is not supported, only 1-6, command is ignored");
        } else if (tmp < 1) {
            if (echo) {
                if (queueFilter.getSortDescending() > 0) {
                    System.out.println("Ascending sort is turned off, will go back to descending sort on column " + queueFilter.getSortDescending());
                } else {
                    System.out.println("Ascending sort is turned off, will show default listing");
                }
            }
            queueFilter.setSortAscending(0);
        } else {
            queueFilter.setSortAscending(tmp);
            queueFilter.setSortDescending(0);
            if (echo)
                System.out.println("Sorting by column " + tmp + " in ascending order");
        }
    }

    private static void setQueuenameFilter(String input) {
        queueFilter.setQueueNames(null);
        if (input.length() > 1) {
            queueFilter.setQueueNames(input.substring(1).trim());
            if (queueFilter.getQueueNames().isEmpty())
                queueFilter.setQueueNames(null);
        }
        if (queueFilter.getQueueNames() == null) {
            if (echo)
                System.out.println("Filtering by queue names is turned off, will match ALL queue listing");

        } else {
            filterHistory.add(input);
            if (echo)
                System.out.println("Queue names filter is set to " + queueFilter.getQueueNames());
        }
        filterChangedViewNotUpdated = true;
    }

    private static void setSearchFilter(String input) {
        queueFilter.setSearchString(null);
        if (input.length() > 1) {
            queueFilter.setSearchString(input.substring(1).trim());
            if (queueFilter.getSearchString().isEmpty())
                queueFilter.setSearchString(null);
        }
        if (queueFilter.getSearchString() == null) {
            if (echo)
                System.out.println("Filtering by search string is turned off, will match ALL queue listing");
        } else {
            filterHistory.add(input);
            if (echo)
                System.out.println("Search string filter is set to '" + queueFilter.getSearchString() + "' (will be negated: " + queueFilter.getSearchString().startsWith("!") + ")");
        }
        filterChangedViewNotUpdated = true;
    }

    private static List<QueueInfo> refreshQueueList(Config config) throws NoSuchAlgorithmException, IOException, KeyManagementException {
        List<QueueInfo> qiList;
        qiList = ListQueues.Retrieve.retrieveAndFilter(config, queueFilter); // updates the qiList and prints it
        ListQueues.Print.printToConsole(qiList, echo);
        filterChangedViewNotUpdated = false;
        return qiList;
    }

    private static void filterHistory() {
        for (int i = 0; i < filterHistory.size(); i++) {
            System.out.println((i + 1) + ": " + filterHistory.get(i));
        }
    }

    private static void setConsumerFilter(String input) {
        Integer tmp = Div.getInt(input);
        if (tmp == null)
            return;
        if (echo) {
            if (tmp < 0)
                System.out.println("Filtering by consumer number is turned off, will match ALL consumer count");
            else
                System.out.println("Consumer count filter is set to " + tmp);
        }
        filterChangedViewNotUpdated = true;
        queueFilter.setConsumerCount(tmp);
    }

    private static Map<String, Login> parseLoginFile(File loginFile) throws IOException {
        // The loginFile has the following format, the 3 last fields are optional:
        // hostname [user password protocol]
        Map<String, Login> loginMap = new HashMap<>();
        for (String line : Files.readAllLines(loginFile.toPath())) {
            if (line.startsWith("#") || line.trim().isEmpty())
                continue;
            String[] parts = line.split(" ");
            if (parts.length == 4) {
                loginMap.put(parts[0], new Login(parts[0], parts[1], parts[2], parts[3]));
            } else if (parts.length == 1) {
                loginMap.put(parts[0], new Login(parts[0], null, null, null));
            }
        }
        return loginMap;
    }


    // Other methods - not directly related to the main method

    private static void printQueueInfoToJsonFile(List<QueueInfo> qiList, String input) throws IOException {
        String filename = input.substring(1).trim();
        if (filename.isEmpty()) {
            System.out.println("No filename specified, command ignored");
            return;
        }
        File file = Div.getValidFileReferenceForWrite(filename);
        if (file != null) {
            ListQueues.Print.printToJson(qiList, file);
        }
    }

    private static void showSettings() {
        String sortSetting = "Not specified";
        if (queueFilter.getSortAscending() > 0) {
            sortSetting = "Ascending by column " + queueFilter.getSortAscending();
        } else if (queueFilter.getSortDescending() > 0) {
            sortSetting = "Descending by column " + queueFilter.getSortDescending();
        }
        String consumerFilter = queueFilter.getConsumerCount() < 0 ? "<ALL>" : String.valueOf(queueFilter.getConsumerCount());
        String sizeFilter = queueFilter.getMinQueueSize() < 1 ? "<ALL>" : String.valueOf(queueFilter.getMinQueueSize());
        String queueNamesFilter = queueFilter.getQueueNames() == null ? "<ALL>" : queueFilter.getQueueNames();
        String searchStringFilter = queueFilter.getSearchString() == null ? "<ALL>" : prettyprinted(queueFilter.getSearchString());
        String limitSetting = queueFilter.getMaxQueueListSize() < 1 ? "<ALL>" : "First " + queueFilter.getMaxQueueListSize() + " rows";
        String ageSetting = messageFilter.getMinAgeMinutes() < 1 ? "<ALL>" : "Older than " + messageFilter.getMinAgeMinutes() + " minutes";
        String typeSetting = messageFilter.getMessageType() == null ? "<ALL>" : prettyprinted(messageFilter.getMessageType());
        String baMessageIdSetting = messageFilter.getBaMessageId() == null ? "<ALL>" : prettyprinted(messageFilter.getBaMessageId());
        String senderSetting = messageFilter.getSender() == null ? "<ALL>" : prettyprinted(messageFilter.getSender());
        String receiverSetting = messageFilter.getReceiver() == null ? "<ALL>" : prettyprinted(messageFilter.getReceiver());


        System.out.println("Que/Msg   Type   Operation        Cmd    Reset Comp    : Setting");
        System.out.println("----------------------------------------------------------------");
        System.out.println("Queue     Sort   Column-sort      (a/d)  (0)   N/A     : " + sortSetting);
        System.out.println("Queue     Filter No of consumers  (c)    (-1)  (==)    : " + consumerFilter);
        System.out.println("Queue     Filter Queue-size       (z)    (0)   (>=)    : " + sizeFilter);
        System.out.println("Queue     Filter Queues (list)    (q)    ()    (or)    : " + queueNamesFilter);
        System.out.println("Queue     Filter Search (regexp)  (s)    ()    (match) : " + searchStringFilter);
        System.out.println("Queue     Filter Limit-N-rows     (l)    (0)   (>=)    : " + limitSetting);
        System.out.println("Message   Filter TTL (minutes)    (age)  (0)   (>)     : " + ageSetting);
        System.out.println("Message   Filter Msg/BusinessType (type) ()    (match) : " + typeSetting);
        System.out.println("Message   Filter BAMessageId      (baid) ()    (match) : " + baMessageIdSetting);
        System.out.println("Message   Filter Sender           (snd)  ()    (match) : " + senderSetting);
        System.out.println("Message   Filter Receiver         (rcv)  ()    (match) : " + receiverSetting);
    }

    private static String prettyprinted(String searchString) {
        return searchString.startsWith("!") ? searchString.substring(1) + " (negated:true)" : searchString + " (negated:false)";
    }


    // Interactive help
    private static void help() {
        System.out.println(" MISC:");
        System.out.println("   h            - help");
        System.out.println("   x            - exit");
        System.out.println("   e<on|off>    - off: will output only the most necessary information, useful for script-mode");
        System.out.println(" LISTINGS:");
        System.out.println("   v            - view queues (refresh list - subsequent purge/delete will be based on this list)");
        System.out.println("   m[<number>]  - view messages in all queues. If number is given, only messages from that queue will be shown");
        System.out.println("   f            - show filter (search/queues) history");
        System.out.println("   y            - show filter/sort settings");
        System.out.println(" QUEUE FILTERS:");
        System.out.println("   c<number>    - lists queues with exactly this number of consumers. Remove filter by setting to -1");
        System.out.println("   z<number>    - lists queues with size larger than or equal to this number.");
        System.out.println("   s<string>    - match lines with this regexp/string. Prefixing with ! will negate search.");
        System.out.println("                  example of searches:");
        System.out.println("                    test            - match rows containing 'test'");
        System.out.println("                    !endpoint|edx   - match rows NOT containing 'endpoint' nor 'edx'");
        System.out.println("                    apt09.*ecp      - match rows containing 'apt09<zero-or-more-char>ecp'");
        System.out.println("   q<queues>    - lists queues with these names (comma-separated). Remove filter by setting q without argument.");
        System.out.println("   a<number>    - ascending sort by column 1-host 2-name 3-size 4-consumers 5-enqueue 6-dequeue. Remove sort by setting to 0.");
        System.out.println("   d<number>    - descending sort by column 1-host 2-name 3-size 4-consumers 5-enqueue 6-dequeue. Remove sort by setting to 0.");
        System.out.println("   l<number>    - limit output to this number of lines (after filter/sorting is applied). Remove limit by setting to 0");
        System.out.println(" MESSAGE FILTERS: These filters are AND'ed together");
        System.out.println("   age<number>  - list messages older than this <number> of minutes. Remove filter by setting to 0");
        System.out.println("   4 ECP/EDX special filters:");
        System.out.println("      - Use % for wildcard, typically %my-search-term%, but can be used anywhere in the string");
        System.out.println("      - Start with ! to negate the filter");
        System.out.println("      - Remove filter by setting to empty string");
        System.out.println("   type<string> - list messages with this message/businesstype. Example: 'typeNBM-AOF-1'");
        System.out.println("   baid<string> - list messages with this BAMessageId. Example: 'baid123%456'");
        System.out.println("   snd<string>  - list messages with this sender. Example: 'snd!50V00000ABC'");
        System.out.println("   rcv<string>  - list messages with this receiver. Example: 'rcv50V00000ABC'");
        System.out.println(" ACTIONS - TAKE CARE!:");
        System.out.println("   p<number>    - purge queue (number is queue number). Purge ALL queues in the printed list if number is -1");
        System.out.println("   r<number>    - same operation as purge (p), but will also apply message filter (-t) to the messages in the queue");
        System.out.println("   n<number>    - nuke/destroy queues (number is queue number). Delete ALL queues in the printed list if number is -1");
        System.out.println("   j<file>      - print queue-info to file in JSON-format");
    }

    // Command line help
    private static void usage() {
        System.out.println("BrokerShell (BRS) " + VERSION + " is made to browse and manage queues on Artemis brokers used by");
        System.out.println("ECP/EDX-installations. The broker must respond on 8161-port (console-api-port) for this tool to work.");
        System.out.println("You can run BRS locally on same host, then things will most likely work. If you want to allow for");
        System.out.println("remote access (which will enable the greatest power of BRS), then change the jolokia-access.xml to this:");
        System.out.println("    <restrict><cors><allow-origin>*</allow-origin></cors></restrict>");
        System.out.println("If you allow for remote access, then you can specify multiple hosts in the -h option, and BRS will list");
        System.out.println("queues and perform delete/purge operations on all hosts listed. Combined with the -f option + crontab");
        System.out.println("you can automate management of queues on multiple brokers from a single command-line tool.");
        System.out.println();
        System.out.println("Usage  : java -jar ekit.jar BRS [OPTIONS] BROKER\n");
        System.out.println("\n\nThe options and arguments:");
        System.out.println(" OPTIONS       : ");
        System.out.println("                 -f <commands-file> : each line in file specify an interactive command.");
        System.out.println("                 -c <commands>      : a semi-colon separated list of interactive commands. Ignored if -f is specified");
        System.out.println("                 -h <host-file>     : each line specify: hostname [user password protocol] (3 last fields are optional)");
        System.out.println(" BROKER        : <URL|PATH> - either specify a broker-URL, or a path for the ECP/EDX-configuration");
        System.out.println("                 URL                : Specify url on this form: http[s]://user:pass@host:port");
        System.out.println("                 PATH               : Specify either the conf-path for ECP or EDX (ex: /etc/ecp-endpoint or /etc/edx-toolbox)");
        System.out.println();
        System.out.println("\nExample 1:  Start BRS in interactive mode");
        System.out.println("\tjava -jar ekit.jar BRS https://endpoint:password@localhost:8161");
        System.out.println("\nExample 2:  Interactive mode - browse and manage queues on multiple hosts");
        System.out.println("\tjava -jar ekit.jar BRS -i -h hosts.txt https://endpoint:password@remotehost:8161");
        System.out.println("\nExample 3:  Script mode - run commands from file");
        System.out.println("\tjava -jar ekit.jar BRS -f commands.txt https://endpoint:password@localhost:8161");
        System.out.println("\nExample 4:  Script mode - find queues with min 1 msg on queue, then find msg 10min or older and remove those");
        System.out.println("\tjava -jar ekit.jar BRS -c \"eoff;z1;t10;v;r-1\" https://endpoint:password@localhost:8161");
    }

}
